/*
 * Tuxcall -- Call waiting detector for Linux
 * Copyright (C) 2001-2002 Angel Luis Jimenez Martinez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#   include <config.h>
#endif

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "iir.h"
#include "vec.h"
#include "common.h"
#include "detect.h"

#define myabs(x) (x>=0 ? x :-x)

extern float LimitMult;
extern int ChunkSize;
extern int CheckPPP;

/*
 * Handles state change (European telephonic system)
 * 
 * Call waiting tone:
 * It consists of two beeps per tone (each about 160 ms), and a gap between them
 * (160 ms too)
 *
 * There are 4000 ms of silence between tones.
 *
 * States:
 *    _____   _____
 *    |   |   |   |
 *  0 | 1 | 2 | 3 | 4
 * ___|   |___|   |___
 *
 */
void stateDetectEU( int * state, unsigned long * pulses, float * totalAcum
    , float chunkMean, int nChunk, int chunkSize, void (*onCall)() )
{
    float totalMean, limit;

    (*totalAcum)+= chunkMean;
    totalMean= (*totalAcum)/nChunk;
    limit= totalMean * LimitMult;

#ifdef DEBUG
    printf( "%lu: cMean=%f, tMean=%f, limit=%f\n"
        , (unsigned long)(((nChunk-1)*chunkSize*1000)/SPEED), chunkMean
        , totalMean, limit );
#endif                    


    switch( *state )
    {
        case 0:
            if ( chunkMean >= limit )
            {
#ifdef DEBUG
                printf(_("%lu - state 0 -> 1 - pulses %lu\n")
                    , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                    , *pulses);
#endif                                
                *state= 1;
                *pulses= 1;
            }    
            break;
        case 1:
            if ( chunkMean >= limit )
                (*pulses)++;
            else
            {
                if ( *pulses>=1 && *pulses<=3 )
                {
#ifdef DEBUG
                    printf(_("%lu - state 1 -> 2 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 2;
                }    
                else
                {
#ifdef DEBUG
                    printf(_("%lu - state 1 -> 0 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 0;
                }
                *pulses= 1;
            }    
            break;
        case 2:
            if ( chunkMean >= limit )
            {
                if (*pulses>=1 && *pulses<=3)
                {
#ifdef DEBUG
                    printf(_("%lu - state 2 -> 3 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 3;
                }    
                else
                {
#ifdef DEBUG
                    printf(_("%lu - state 2 -> 1 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 1;
                }
                *pulses= 1;
            }    
            else
            {
                (*pulses)++;
            }    
            break;
        case 3:
            if ( chunkMean >= limit )
                (*pulses)++;
            else
            {
                if (*pulses>=1 && *pulses<=3)
                {
#ifdef DEBUG
                    printf(_("%lu - state 3 -> 4 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 4;
                }    
                else
                {
#ifdef DEBUG
                    printf(_("%lu - state 3 -> 0 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 0;
                }
                *pulses= 1;
            }    
            break;
        case 4:
#ifdef DEBUG
            printf(_("%lu - CALL DETECTED!\n")
                , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED));
#endif                            
            fprintf(stderr, _("%lu - CALL DETECTED!\n")
                , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED));

            if ( onCall != NULL )    
                (*onCall)();    
    
            *state= 0;
            *pulses= 1;
    }
}

/*
 * Handles state change (USA telephonic system)
 * 
 * Call waiting tone:
 * It consists of one beep per tone (300 ms), and 10 seconds of silence 
 * between tones.
 *
 * States:
 *    _____   
 *    |   |   
 *  0 | 1 | 2    and 3 to ensure state 2 is long enough
 * ___|   |___
 *
 */
void stateDetectUSA( int * state, unsigned long * pulses, float * totalAcum
    , float chunkMean, int nChunk, int chunkSize, void (*onCall)() )
{
    float totalMean, limit;

    (*totalAcum)+= chunkMean;
    totalMean= (*totalAcum)/nChunk;
    limit= totalMean * LimitMult;

#ifdef DEBUG
    printf( "%lu: cMean=%f, tMean=%f, limit=%f\n"
        , (unsigned long)(((nChunk-1)*chunkSize*1000)/SPEED), chunkMean
        , totalMean, limit );
#endif                    


    switch( *state )
    {
        case 0:
            if ( chunkMean >= limit )
            {
#ifdef DEBUG
                printf(_("%lu - state 0 -> 1 - pulses %lu\n")
                    , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                    , *pulses);
#endif                                
                *state= 1;
                *pulses= 1;
            }    
            break;
        case 1:
            if ( chunkMean >= limit )
                (*pulses)++;
            else
            {
                if ( *pulses>=4 && *pulses<=6 )
                {
#ifdef DEBUG
                    printf(_("%lu - state 1 -> 2 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 2;
                }    
                else
                {
#ifdef DEBUG
                    printf(_("%lu - state 1 -> 0 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 0;
                }
                *pulses= 1;
            }    
            break;
        case 2:
            if ( chunkMean >= limit )
            {
                if (*pulses>=5) /* TODO: Change number???*/
                {
#ifdef DEBUG
                    printf(_("%lu - state 2 -> 3 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 3;
                }    
                else
                {
#ifdef DEBUG
                    printf(_("%lu - state 2 -> 0 - pulses %lu\n")
                        , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED)
                        , *pulses);
#endif                                
                    *state= 0;
                }
                *pulses= 1;
            }    
            else
            {
                if ( *pulses>= 5 )
                    *state= 3;
                (*pulses)++;
            }    
            break;
        case 3:
#ifdef DEBUG
            printf(_("%lu - CALL DETECTED!\n")
                , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED));
#endif                            
            fprintf(stderr, _("%lu - CALL DETECTED!\n")
                , (unsigned long)(((float)nChunk*chunkSize*1000)/SPEED));

            if ( onCall != NULL )    
                (*onCall)();    
    
            *state= 0;
            *pulses= 1;
    }
}

/*
 * It does a step of the detection...
 * 
 * Read max readLen smaples from buff and put filtered data on filtered
 * using the coefficients from a & b for the IIR filtering
 * stateDetect: It does all the state change/detection. This way other tones 
 * could be identified.
 * If "afterChunk" != NULL executes that function after processing each data 
 # chunk
 * If "onCall" != NULL executes that function when a call is detected
 * 
 * returns false if ppp connection has finished (i.e. ppp is down)
 */
int detectStep( signed short * buff, float * filtered, long readLen
            , float * a, float * b
            , void (*stateDetect)( int * state, unsigned long * pulses
                                    , float * totalAcum, float chunkMean
                                    , int nChunk, int chunkSize
                                    , void (*onCall)() )
            , void (*afterChunk)(float, float)
            , void (*onCall)() )
{
    static int i, state= 0;
    static float chunkAcum=0, chunkMean= 0, totalAcum= 0;
    static long cont2= 0;
    static unsigned long pulses= 0, cont= 0, nChunk= 0;

    recursiveNarrow( buff, filtered, readLen, a, b );

    for ( i= 0; i<readLen; i++ )
    {
        if( (cont2) == ChunkSize )
        {
            cont2= 0;
            nChunk++;

            chunkMean= chunkAcum/ChunkSize;

            /* Every x chunks test if ppp connection is up */
            if ( (nChunk % 32) == 0 )
            {
                if ( CheckPPP && !connected())
                    return 0;
            }

            (*stateDetect)( &state, &pulses, &totalAcum, chunkMean, nChunk
                        , ChunkSize, onCall );

            if ( afterChunk != NULL )
            {
                (*afterChunk)(chunkMean, (totalAcum/nChunk)*LimitMult );
            }    

            chunkAcum= 0;            
        }

        chunkAcum+= myabs( filtered[i] ); 
        cont2++;
    }    
    cont+= readLen;

    filtered[0]= filtered[readLen - 2];
    filtered[1]= filtered[readLen - 1];

    return 1;
}        

